/* ----------------------------------------------------------------------------
 *
 * Raima Database Manager
 *
 * Copyright (c) 2012 Raima Inc.,  All rights reserved.
 *
 * Use of this software, whether in source code format, or in executable,
 * binary object code form, is governed by the Raima LICENSE which
 * is fully described in the LICENSE.TXT file, included within this
 * distribution of files.
 *
 * --------------------------------------------------------------------------*/
#include <stdio.h>
#include <iostream>

/* rdm.h is required in all files using the RDM Database */
#include "rdm.h"

/* perf01.h was generated from the DDL file perf01.sdl. */
#include "perf01.h"

/* perf01_dbd.h was generated from the DDL file perf01.sdl. using rdmsql with the -c option */
#include "perf01_dbd.h"

/* perf01_api.h was generated from the DDL file perf01.sdl. using rdmsql with the -x option */
#include "perf01_api.h"

/* Name space for the RDM C++ interface */
using namespace RDM_CPP;

/*! \file perf01.cpp
 * \brief Source code for the RDM Perf01 example using the RDM C++ interface
 */
 
/*!
 * \mainpage Perf01CPP
 * \author Jeffrey R. Parsons, Raima Sr. Engineer
 * \date 2012
 *
 * \section Description
 * EXAMPLE - Perf01CPP
 *
 * This version of Perf01 example performs a number of atomic database 
 * operations on a very simple database schema using the RDM C++ Interface.
 *
 * Each time you run this example
 *   * The database is initialized (removing all existing data)
 *   * 50,000 new records are created
 *   * Each record is read
 *   * Each record is updated
 *   * Each record is read again
 *   * Each record is deleted
 *
 * \section Perf01 Database
 * The DDL (Database Definition Language) specification for the database used
 * in this example is located in the file perf01.sdl. The example utilizes 
 * the SQL DDL syntax and specifies an inmemory database with a single 
 * table (simple).  The table has a single column (int_col).  There are no 
 * index definitions for this database.
 *
 * \include perf01.sdl
 *
 * The schema was compiled using the RDM rdmsql utility with the -c and -x option to
 * generate an embedded database dictionary (used by the RDM d_open_ptr API) and
 * the C++ interface.
 *
 * \code rdmsql -c -x perf01.sdl \endcode
 *
 * The database dictionary information is embedded inside the application so 
 * an external DBD file stored on the file system is not required.
 *
 * \section RDM C++ Interface Methods Used
 * The following RDM APIs and Generated C++ methods are illustrated:
 *   * d_tfsinitEx/d_tfsterm
 *   * Dbperf01.Open
 *   * Dbperf01.New_simple_recordWithFieldValues
 *   * Dbperf01.Get_simple_records
 *   * Cursors_simple.AfterLast
 *   * Cursor_simple++
 *   * Cursor_simple.Set_int_col
 *   * Cursor_simple.Last
 *   * Cursors_simple.BeforeFirst
 *   * Cursor_simple--
 *   * Cursor_simple.Delete
 *
 * They are fully documented in http://docs.raima.com/rdm/.
 */

class perf01 {
public:
    Db_perf01 db;
    TFS_HANDLE hTFS;

    void Initialize () throw (rdm_exception&);
    void Cleanup () throw (rdm_exception&);
    void AddRecords() throw (rdm_exception&);
    void ReadRecords() throw (rdm_exception&);
    void UpdateRecords() throw (rdm_exception&);
    void DeleteRecords() throw (rdm_exception&);
    void DisplayElapsedTime(
        const char *msg,         /**< [in] Message to be display along with the elapsed time */
        uint64_t    start_time,  /**< [in] Start of elapsed time to display */
        uint64_t    end_time);   /**< [in] End of elapsed */
};

/*! \brief Initialize the RDM TFS and initialize the perf01 database
 *
 * This method initializes the RDM Transactional File Server (TFS) to use 
 * the single-user non-transactional STANDALONE implementation.  It also opens
 * the perf01 databause via the generated Db_perf01 class in exclusive ('x')
 * mode.  Exclusive mode does not require database locks or transactions.
 */
void perf01::Initialize()
{
    int32_t  rdm_status = S_OKAY;
    uint64_t start_time;
    uint64_t end_time;

    /* Initialize the TFS to use the STANDALONE configuration (TFSS) */
    start_time = psp_timeNanoSecs();
    rdm_status = d_tfsinitEx(NULL, NULL, &hTFS, TFS_TYPE_STANDALONE);
    if(rdm_status != S_OKAY)
    {
        cerr << "Error initializing the TFS (" << d_errorname(rdm_status) << " - " << d_errorstr(rdm_status) << ")" << endl;
        throw (rdm_exception(rdm_status));
        return;
    }

    /* Open and initialize the perf01 database */
    try
    {
        db = db.Open("hwcpp", "x");
        db.Initialize();
    }
    catch (rdm_exception& e)
    {
        cerr << "Error " << e.error() << " (" << e.what() << ") performing perf01 database initialization" << endl;
    }
    end_time = psp_timeNanoSecs();  
    
    /* Display information */
    cout << "Initializing with STANDALONE configuration" << endl;
    DisplayElapsedTime("Preparing a new database:\t\t\t", start_time, end_time);
    cout << "\tOne, simple record type." << endl;
    cout << "\tIn-memory" << endl;
    cout << "\tOpened exclusively" << endl;

    return;
}

/*! \brief Terminate the RDM TFS
 *
 * This method terminates the STANDALONE TFS used
 * in the example. 
 *
 */
void perf01::Cleanup()   /**< [out] Pointer to the RDM task to be cleaned up */
{
    uint64_t start_time;
    uint64_t end_time;

    start_time = psp_timeNanoSecs();

    /* Terminate the STANDALONE TFS */
    d_tfsterm(hTFS);

    end_time = psp_timeNanoSecs();

    DisplayElapsedTime("Cleanup:\t\t\t\t\t", start_time, end_time);

    return;
}

/*! \brief Display a message and the elapsed time to perform database operations
 *
 */
void perf01::DisplayElapsedTime(
    const char *msg,
    uint64_t    start_time,
    uint64_t    end_time)
{
    uint64_t elapsed_time = end_time - start_time;
    cout << msg << elapsed_time / 1000000 << " milliseconds" << endl;
}

/*! \brief Insert 50,000 simple records into the perf01 database
 *
 * This function adds 50,000 simple records into the inmemory perf database.  The
 * simple record contains only a single integer field with no index.  The function will
 * calculate and display the total time elapsed while inserting the records
 */
void perf01::AddRecords()
{
    int32_t       ii;
    struct simple simple_rec;
    Cursor_simple simple_input;
    uint64_t      start_time;
    uint64_t      end_time;

    memset(&simple_rec, 0, sizeof(simple_rec));

    start_time = psp_timeNanoSecs();    
    try
    {
        /* Loop to add 50,000 new simple record */
        for(ii=0; ii<50000; ii++)
        {
            /* Set values for the new record */
            simple_rec.int_col = 1;

            /* Add the record to the database */
            simple_input = db.New_simple_recordWithFieldValues(&simple_rec);                
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error addming a simple record to the perf01 database (" << e.error() << " - " << e.what() << ")" << endl;
        throw (e);
        return;
    }
    end_time = psp_timeNanoSecs();

    DisplayElapsedTime("Add 50,000 records:\t\t\t\t", start_time, end_time);
    return;
}

/*! \brief Read all simple records and calculate the sum of the int_col column
 *
 * This function gets a cursor to iterate through and reads all simple records
 * stored in the perf01 database.  The iteration is done in raw file order 
 * through the cursor created by the generated Db->Get_simple_records APIs.
 * The function will calculate the sum of the int_col field from all simple
 * records in the database.  The function will also calculate and display the
 * total time elapsed while reading the records.
 */
void perf01::ReadRecords()
{
    int32_t       rdm_status = S_OKAY;
    uint64_t      total = 0;
    struct simple simple_rec;
    uint64_t      start_time;
    uint64_t      end_time;
    Cursor_simple simple_cursor;

    memset(&simple_rec, 0, sizeof(simple_rec));

    start_time = psp_timeNanoSecs();    
    try
    {
        /* Loop through each simple record in file order */
        simple_cursor = db.Get_simple_records();
        while(!simple_cursor.AfterLast())
        {
            /* Read the record */
            simple_cursor.GetFieldValues(&simple_rec);
            
            /* Add the value read to the running sum */
            total += simple_rec.int_col;

            /* Go to the next record */
            simple_cursor++;
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error reading simple records from the perf01 database (" << e.error() << " - " << e.what() << ")" << endl;
        throw (e);
        return;
    }
    end_time = psp_timeNanoSecs();

    DisplayElapsedTime("Scanning, reading, summing 50,000 records:\t", start_time, end_time);
    cout << "\tSum = " << total << endl;
    return;
}

/*! \brief Update all simple records in the database
 *
 * This function iterates through and updates all simple records stored in the
 * perf01 database.  The iteration is done in raw file order through a cursor
 * obtained by the generated db->Get_simple_records method.  The function will
 * read, increment, and then update the value stored in the int_col field.
 * The function will calculate and display the total time elapsed while 
 * deleting the records.
 */
void perf01::UpdateRecords()
{
    int32_t       rdm_status = S_OKAY;
    struct simple simple_rec;
    Cursor_simple simple_cursor;
    uint64_t      start_time;
    uint64_t      end_time;

    memset(&simple_rec, 0, sizeof(simple_rec));

    start_time = psp_timeNanoSecs();    
    try
    {
        /* Loop through each simple record in file order */
        simple_cursor = db.Get_simple_records();
        while(!simple_cursor.AfterLast())
        {
            /* Read the record */
            simple_cursor.GetFieldValues(&simple_rec);
            
            /* increment the int_col value */
            simple_rec.int_col++;

            /* Update the field in the database */
            simple_cursor.Set_int_col(simple_rec.int_col);

            /* Go to the next record */
            simple_cursor++;
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error updating simple records in the perf01 database (" << e.error() << " - " << e.what() << ")" << endl;
        throw (e);
        return;
    }
    end_time = psp_timeNanoSecs();
    
    DisplayElapsedTime("Updating 50,000 records:\t\t\t", start_time, end_time);
    return;
}

/*! \brief Delete all simple records from the database
 *
 * This function will iterate and delete all simple records stored in the 
 * perf01 database.  The iteration is done in raw file order through a cursor
 * obtained by the generated db->Get_simple_records method.  Unlike the
 * previous function this implementation starts with the last record in
 * the file and interates to the first record in the file.  The function will 
 * calculate and display the total time elapsed while deleting the records.
 */
void perf01::DeleteRecords()
{
    int32_t       rdm_status = S_OKAY;
    uint64_t      start_time;
    uint64_t      end_time;
    Cursor_simple simple_cursor;

    start_time = psp_timeNanoSecs();
    try
    {
        /* Loop through each simple record in file order (last to first) */
        simple_cursor = db.Get_simple_records().Last();
        while(!simple_cursor.BeforeFirst())
        {
            /* Delete the record */
            simple_cursor.Delete();

            /* Go to the previous record */
            simple_cursor--;
        }
    }
    catch (rdm_exception& e)
    {
        cerr << "Error deleting simple records from the perf01 database (" << e.error() << " - " << e.what() << ")" << endl;
        throw (e);
        return;
    }    
    end_time = psp_timeNanoSecs();


    DisplayElapsedTime("Deleting 50,000 records:\t\t\t", start_time, end_time);
    return;
}

/*! \brief Main function for perf01 example
 *
 * The function initializes the perf01 cleass environment and runs the 
 * AddRecords, ReadRecords, UpdateRecords, and DeleteRecords methods.
 *
 * @return Returns 0 on success, and 1 if there were any errors
 */
int32_t main(void)
{
    perf01      perf;

    /* Display the example title */
    cout << RDM_EXAMPLE_DESC("Perf01 C++");

    try
    {
        /* Initialize the database related items in the class */
        perf.Initialize();

        /* Add 50,000 records to the database */
        perf.AddRecords();

        /* Read all of the records in the database */
        perf.ReadRecords();

        /* Update all of the records in the database */
        perf.UpdateRecords();

        /* Read all of the records in the database */
        perf.ReadRecords();

        /* Remove all of the records in the database */
        perf.DeleteRecords();

    }
    catch (rdm_exception& e)
    {
        perf.Cleanup();
        cerr << "Example application aborting due to an internal error (" << e.error() << " - " << e.what() << ")" << endl;
        return 1;
    }

    cout << endl << "Press Enter to terminate application";
    getchar();
	return 0;
}